import hashlib
import json
import logging
import random
import time
from typing import Optional, Text, Dict, List, Any, Union

from sanic_jwt import exceptions
from sqlalchemy import and_
from sqlalchemy.orm import Session

from rasax.community import config
from rasax.community.database.admin import User, Role, SingleUseToken

logger = logging.getLogger(__name__)


class AuthMechanisms:
    username_password = "username/password"
    saml = "saml"


class MismatchedPasswordsException(Exception):
    def __init__(self):
        self.message = "Passwords do not match!"

    def __str__(self):
        return self.message


class UserException(Exception):
    def __init__(self, username):
        self.message = username

    def __str__(self):
        return self.message


class RoleException(Exception):
    def __init__(self, username):
        self.message = username

    def __str__(self):
        return self.message


class UserService(object):
    def __init__(self, session: Session):
        self.session = session

    def fetch_user(
        self, username: Text, return_api_token: bool = False
    ) -> Optional[Dict]:
        user = self._fetch_user(username)
        if user:
            return user.as_dict(return_api_token=return_api_token)
        else:
            return None

    def _fetch_user(self, username: Text) -> User:
        return self.session.query(User).filter(User.username == username).first()

    def fetch_all_users(
        self, team: Text, exclude_system_user: bool = True
    ) -> List[Dict]:
        query = User.team == team

        if exclude_system_user:
            query = and_(query, User.username != config.SYSTEM_USER)

        users = self.session.query(User).filter(query).all()
        return [u.as_dict() for u in users]

    def create_user(
        self,
        username: Text,
        raw_password: Text,
        team: Text,
        roles: Optional[Union[List[Text], Text]],
        auth_mechanism: Text = AuthMechanisms.username_password,
    ):
        existing_user = self._fetch_user(username)
        if existing_user:
            raise UserException(username)
        password = self.hash_pw(raw_password)
        api_token = self.generate_api_token()
        new_user = User(
            username=username,
            password_hash=password,
            project="default",
            team=team,
            api_token=api_token,
            authentication_mechanism=auth_mechanism,
        )
        self.session.add(new_user)
        self.session.commit()

        if isinstance(roles, str):
            roles = [roles]

        for role in roles:
            self.add_role_to_user(username, role)

    def delete_user(self, username: Text):
        existing_user = self._fetch_user(username)
        if not existing_user:
            raise UserException(username)

        if username == config.SYSTEM_USER:
            raise UserException("Cannot delete user '{}'.".format(username))

        deleted_user = existing_user.as_dict()
        self.session.delete(existing_user)
        self.session.commit()

        return deleted_user

    def add_role_to_user(self, username: Text, role: Text) -> None:
        """Gives the user an additional role."""

        existing_user = self._fetch_user(username)

        if not existing_user:
            raise UserException("User '{}' does not exist.".format(username))

        _role = self.session.query(Role).filter(Role.role == role).first()
        if not _role:
            raise RoleException("Role '{}' does not exist.".format(role))

        if not any([r.role == _role.role for r in existing_user.roles]):
            existing_user.roles.append(_role)
        else:
            logging.debug(
                "User '{}' already had role '{}'.".format(username, _role.role)
            )

        self.session.commit()

    def replace_user_roles(
        self, username: Text, roles: Optional[Union[List[Text], Text]]
    ) -> None:
        """Removes all roles from a user and replaces them with new ones."""

        existing_user = self._fetch_user(username)

        for _role in existing_user.roles:
            existing_user.roles.remove(_role)

        if roles:
            if not isinstance(roles, list):
                roles = [roles]
            for role in roles:
                self.add_role_to_user(username, role)

        self.session.commit()

    def delete_user_role(self, username: Text, role: Text) -> None:
        """Deletes a role from a user."""

        existing_user = self._fetch_user(username)
        if not existing_user:
            raise UserException(username)

        roles_to_delete = [r for r in existing_user.roles if r.role == role]

        if not roles_to_delete:
            raise RoleException(role)

        for role in roles_to_delete:
            existing_user.roles.remove(role)

        self.session.commit()

    def create_saml_user(
        self,
        name_id: Text,
        attributes,
        team: Text,
        role: Text,
        username: Optional[Text] = None,
        api_token: Optional[Text] = None,
    ) -> None:
        api_token = api_token or self.generate_api_token()

        existing_name_id = self._fetch_user(name_id)
        if existing_name_id:
            raise UserException(
                "Username for SAML ID '{}' already exists.".format(name_id)
            )

        new_user = User(
            username=username or name_id,
            name_id=name_id,
            project="default",
            team=team,
            api_token=api_token,
            authentication_mechanism=AuthMechanisms.saml,
            attributes=json.dumps(attributes),
            username_is_assigned=username is not None,
        )
        self.session.add(new_user)
        self.session.commit()

        self.add_role_to_user(username if username else name_id, role)

    def update_saml_username(
        self, name_id: Text, username: Text
    ) -> Optional[Dict[str, Any]]:
        """Creates a new user for `username`.

        Deletes the old user associated with `name_id`.
        """

        existing_user = self._fetch_saml_user(name_id)

        if not existing_user:
            raise UserException("User with SAML ID '{}' not found".format(name_id))

        if self._fetch_user(username):
            raise UserException("Username '{}' already exists".format(username))

        # delete old user
        _ = self.delete_user(existing_user.username)

        self.create_saml_user(
            name_id=name_id,
            attributes=json.loads(existing_user.attributes),
            team=existing_user.team,
            role=existing_user.roles[0].role,
            username=username,
            api_token=existing_user.api_token,
        )

        return self.fetch_saml_user(name_id)

    def _fetch_saml_user(self, name_id: Text) -> User:
        return self.session.query(User).filter(User.name_id == name_id).first()

    def fetch_saml_user(self, name_id: Text) -> Optional[Dict[str, Any]]:
        user = self._fetch_saml_user(name_id)
        if user:
            return user.as_dict()
        else:
            return None

    def _delete_single_use_token(self, token: Text) -> None:
        existing_token = (
            self.session.query(SingleUseToken)
            .filter(SingleUseToken.token == token)
            .first()
        )
        if not existing_token:
            return None

        self.session.delete(existing_token)
        self.session.commit()

    def update_single_use_token(
        self, name_id: Text, single_use_token: Text, lifetime: float = 60.0  # 1 minute
    ) -> None:
        existing_name_id_user = (
            self.session.query(User).filter(User.name_id == name_id).first()
        )

        if not existing_name_id_user:
            raise UserException("name_id '{}' not found".format(name_id))

        expires = time.time() + lifetime

        existing_token = existing_name_id_user.single_use_token

        if existing_token:
            self._delete_single_use_token(existing_token.token)

        new_token = SingleUseToken(
            token=single_use_token, expires=expires, username=name_id
        )
        existing_name_id_user.single_use_token = new_token

        self.session.commit()

    def single_use_token_login(
        self, single_use_token: Text, return_api_token: bool = False
    ) -> Dict:
        user = (
            self.session.query(User)
            .filter(User.single_use_token.has(SingleUseToken.token == single_use_token))
            .first()
        )
        if not user:
            logger.debug(
                "No user found for single-use token '{}'.".format(single_use_token)
            )
            return None

        # check if token has expired
        token_expires = user.single_use_token.expires
        if time.time() > token_expires:
            logger.debug(
                "single-use token '{}' expired at '{}'"
                "".format(single_use_token, token_expires)
            )
            return None

        # delete single-use token
        self._delete_single_use_token(single_use_token)

        return user.as_dict(return_api_token=return_api_token)

    @staticmethod
    def hash_pw(pw):
        salted_pw = (config.password_salt + pw).encode()
        return hashlib.sha256(salted_pw).hexdigest()

    @staticmethod
    def generate_api_token():
        body = "{}".format(random.random()).encode()
        return hashlib.sha1(body).hexdigest()

    @staticmethod
    def is_username_password_user(user: User) -> bool:
        return user.authentication_mechanism == AuthMechanisms.username_password

    def login(
        self, username: Text, password: Text, return_api_token: bool = False
    ) -> Dict:
        pw_hash = self.hash_pw(password)
        user = self._fetch_user(username)

        if user is None:
            raise exceptions.AuthenticationFailed("Incorrect user or password.")

        if username == config.SYSTEM_USER:
            raise exceptions.AuthenticationFailed(
                "Cannot log in user '{}'.".format(username)
            )

        if not self.is_username_password_user(user):
            logger.info(
                "Cannot log in user '{}' with username/password. User "
                "has auth mechanism '{}'."
                "".format(user.username, user.authentication_mechanism)
            )
            raise exceptions.AuthenticationFailed("Incorrect user or password.")

        if user.password_hash != pw_hash:
            raise exceptions.AuthenticationFailed("Incorrect user or password.")

        return user.as_dict(return_api_token=return_api_token)

    def change_password(self, fields: Dict[Text, Text]) -> Optional[Dict]:
        username = fields["username"]
        user = self.login(username, fields["old_password"])

        if user is None:
            return None
        elif fields["new_password"] != fields["new_password_confirm"]:
            raise MismatchedPasswordsException

        user = self._fetch_user(username)
        user.password_hash = self.hash_pw(fields["new_password"])

        return user.as_dict()

    def admin_change_password(self, username: Text, password: Text) -> Dict:
        existing_user = self._fetch_user(username)
        if not existing_user:
            raise UserException(username)

        existing_user.password_hash = self.hash_pw(password)

        return existing_user.as_dict()

    def api_token_auth(self, api_token: Text, return_api_token: bool = False) -> Dict:
        user = self.session.query(User).filter(User.api_token == api_token).first()
        if user is None:
            raise exceptions.AuthenticationFailed("Incorrect api_token.")
        return user.as_dict(return_api_token=return_api_token)

    def create_project(self, user: Optional[Dict], project_id: Text) -> None:
        owner = self._fetch_user(user.get("username"))
        owner.project = project_id
        owner.role_name = "admin"


def has_role(user: Dict[Text, Any], role: Text) -> bool:
    """Checks whether the user possesses a role."""

    return role in user.get("roles")
