import logging
from typing import Text, List

from sqlalchemy.orm import Session

import rasax.community.config as rasa_x_config
from rasax.community.database import Project
from rasax.community.services.projects_service import ProjectsService
from rasax.community.services.role_service import RoleService, DEFAULT_ROLES
from rasax.community.services.settings_service import SettingsService
from rasax.community.services.user_service import UserService

logger = logging.getLogger(__name__)


def create_initial_project(session):
    if not session.query(Project).first():
        settings_service = SettingsService(session)
        settings_service.init_project(
            rasa_x_config.team_name,
            rasa_x_config.project_name,
            rasa_x_config.LOCAL_MODE,
        )

        session.commit()
        logger.debug(
            "No projects present. Created initial default project '{}'."
            "".format(rasa_x_config.project_name)
        )


def create_default_roles(session):
    role_service = RoleService(session)
    existing_roles = role_service.roles

    for role in [*DEFAULT_ROLES]:
        if role not in existing_roles:
            role_service.save_role(role)
            logger.debug("Created role '{}'.".format(role))
    session.commit()


def create_default_permissions(session):
    role_service = RoleService(session)

    default_roles = role_service.default_roles.items()
    for role, permissions in default_roles:
        if not role_service.get_role_permissions(role):
            role_service.save_permissions_for_role(role, permissions)
            logger.debug("Created default permissions for '{}' role.".format(role))
    session.commit()


def generate_chat_token(session):
    project_service = ProjectsService(session)
    existing_token = project_service.get_token()
    if not existing_token:
        generated_token = project_service.generate_and_save_token()
        logger.debug(
            "Generated chat token '{}' with expiry date {}"
            "".format(generated_token.token, generated_token.expires)
        )


def create_system_user(session):
    user_service = UserService(session)
    if user_service.fetch_user(rasa_x_config.SYSTEM_USER):
        logger.debug(
            "Found existing system system user '{}'.".format(rasa_x_config.SYSTEM_USER)
        )
        return

    user_service.create_user(
        rasa_x_config.SYSTEM_USER, "", rasa_x_config.team_name, "admin"
    )
    logger.debug("Created new system user '{}'.".format(rasa_x_config.SYSTEM_USER))


def _migrate_new_permission(
    session: Session, new_permission: Text, roles_to_migrate: List[Text]
) -> None:
    role_service = RoleService(session)

    for role in roles_to_migrate:
        current_permissions = role_service.get_role_permissions(
            role, rasa_x_config.project_name
        )

        if new_permission not in current_permissions:
            role_service.save_permissions_for_role(role, [new_permission])
            logger.debug(
                "Added permission '{}' for role '{}'.".format(new_permission, role)
            )


def run_migrations(session):
    create_initial_project(session)
    create_default_roles(session)
    create_default_permissions(session)
    create_system_user(session)
    generate_chat_token(session)
